(*
   Windows Porting 99/04/06   by Dojisung
*)

unit WinHan;

interface

uses
   SysUtils, WS2JH;

const
   SHIFT_CODE      = 1000;

type

   THangulUnit     = (huPixel, huGradation);
   THangulMode     = (hmJohap, hmWansung);

   TAsciiFontData  = array[0.. 95,0..15] of byte;
   THangulFontData = array[0..499,0..15,0..1] of byte;

   THangulBuffer   = array[0..33] of byte;

   TPutPixel       = procedure(x, y : integer; Color : Longint);

   THangul         = class

	  constructor Create(ProcAddr : TPutPixel);
	  constructor CreateEx(ProcAddr : TPutPixel; AsciiFile, HangulFile : PChar);
	  procedure   Free;

   private
	  AsciiFontData   : ^TAsciiFontData;
	  HangulFontData  : ^THangulFontData;
	  FHangulUnit     : THangulUnit;
	  FHangulMode     : THangulMode;

   public
	  Gradient    : array[1..3] of byte;
	  PutPixel    : TPutPixel;

	  procedure   Initialize(AsciiFile, HangulFile : PChar);

	  function    GetHangulO(s : String) : String;
	  function    GetHangulRo(s : String) : String;

	  procedure   PrintHangul(x, y : integer; S : string; Color : Longint; IsBold : boolean);

	  property    HangulUnit : THangulUnit read FHangulUnit write FHangulUnit;
	  property    HangulMode : THangulMode read FHangulMode write FHangulMode;
   end;

implementation

uses
   Windows;

const
   SCAN_CODE    = 256;
   BAR_COLOR    = $0000;

   key_table1   : array[0..25] of byte = (
                8,126,16,13,5,7,20,113,105,107,
				103,129,127,120,104,110,9,2,4,11,
                111,19,14,18,119,17);
   key_table2   : array[0..25] of byte = (
                8,126,16,13,6,7,20,113,105,107,
                103,129,127,120,106,112,10,3,4,12,
                111,19,15,18,119,17);
   third_table1 : array[0..20] of byte = (
                0,1,2,3,5,8,1,9,17,19,1,21,22,23,24,1,25,26,27,28,29);
   first_table  : array[0..29] of byte = (
                0,1,2,3,100,4,101,102,5,7,
                103,104,105,106,107,108,109,8,1,9,
                110,11,12,13,14,16,17,18,19,20);
   third_table2 : array[0..10,0..2] of byte = (
                (2,11,4),(5,14,6),(5,20,7),(9,2,10),(9,8,11),(9,9,12),
                (9,11,13),(9,18,14),(9,19,15),(9,20,16),(19,11,20));
   second_table : array[0..6,0..2] of byte = (
				(13,3,14),(13,4,15),(13,29,18),(20,7,21),
                (20,10,22),(20,29,23),(27,29,28));
   init_hangul_code : array[1..3] of byte = (1,2,1);


constructor THangul.Create(ProcAddr : TPutPixel);
begin

   Inherited Create;

   PutPixel    := ProcAddr;
   HangulUnit  := huPixel;
   HangulMode  := hmWansung;
   Gradient[1] := 24;
   Gradient[2] := 24;
   Gradient[3] := 24;

   AsciiFontData  := nil;
   HangulFontData := nil;

end;

constructor THangul.CreateEx(ProcAddr : TPutPixel; AsciiFile, HangulFile : PChar);
begin

   Inherited Create;

   PutPixel   := ProcAddr;
   HangulUnit := huPixel;
   HangulMode := hmWansung;
   Gradient[1] := 24;
   Gradient[2] := 24;
   Gradient[3] := 24;

   AsciiFontData  := nil;
   HangulFontData := nil;

   Initialize(AsciiFile, HangulFile);
end;

procedure   THangul.Free;
begin
   if Assigned(AsciiFontData) then
	  Dispose(AsciiFontData);
   if Assigned(HangulFontData) then
	  Dispose(HangulFontData);

   AsciiFontData  := nil;
   HangulFontData := nil;

   Inherited;
end;

procedure   THangul.Initialize(AsciiFile, HangulFile : PChar);
var
   F          : file;
   s          : string;
   ActualRead : integer;
begin

   if Assigned(AsciiFontData ) then Dispose(AsciiFontData );
   if Assigned(HangulFontData) then Dispose(HangulFontData);

   AsciiFontData  := nil;
   HangulFontData := nil;

   try
	  if AsciiFile <> '' then begin
		 s := AsciiFile;
		 if Pos('.',s) = 0 then s := s + '.Fnt';
		 AssignFile(F,s);
		 Reset(F,1);
		 New(AsciiFontData);
		 BlockRead(F,AsciiFontData^,sizeof(AsciiFontData^),ActualRead);
		 CloseFile(F);
	  end;
   except
   end;

   try
	  if HangulFile <> '' then begin
		 s := HangulFile;
		 if Pos('.',s) = 0 then s := s + '.Fnt';
		 AssignFile(F,s);
		 Reset(F,1);
		 New(HangulFontData);
		 BlockRead(f,HangulFontData^,sizeof(HangulFontData^),ActualRead);
		 CloseFile(F);
	  end;
   except
(*     ó  ޼ ̾α
	  on E : EInOutError do
		 if E.ErrorCode = 2 then MessageDlg('File I/O Error : '+E.Message+' <'+s+'>',mtError,[mbOK],0)
							else MessageDlg('File I/O Error : '+E.Message,           mtError,[mbOK],0);
	  on E : Exception   do
		 MessageDlg('Unexpected Error: '+E.Message,mtError,[mbOK],0);
*)
   end;

   if not Assigned(AsciiFontData ) then AsciiFontData  := nil;
   if not Assigned(HangulFontData) then HangulFontData := nil;

end;

function    THangul.GetHangulO(s : String) : String;
var
   sJH  : String;
begin
   sJH := ConvertWS2JH(s);
   if (Ord(sJH[Length(sJH)]) and $1F) > 1 then begin
	  GetHangulO := s + '';
   end else begin
	  GetHangulO := s + '';
   end;
end;

function    THangul.GetHangulRo(s : String) : String;
var
   sJH  : String;
   Temp : Byte;
begin
   sJH  := ConvertWS2JH(s);
   Temp := Ord(sJH[Length(sJH)]) and $1F;
   // ħ   ħ   ƴϸ
   if (Temp > 1) and (Temp <> 9) then begin
      GetHangulRo := s + '';
   end else begin
      GetHangulRo := s + '';
   end;
end;

procedure   THangul.PrintHangul(x, y : integer; S : string; Color : Longint; IsBold : boolean);
const
   ConvertTable : array[1..3,0..30] of byte = (
      ( 0, 0, 1, 2, 3, 4, 5, 6, 7, 8,
        9,10,11,12,13,14,15,16,17,18,
       19, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 1, 2, 3, 4, 5, 0, 0,
        6, 7, 8, 9,10,11, 0, 0,12,13,
       14,15,16,17, 0, 0,18,19,20,21,0),
      ( 0, 0, 1, 2, 3, 4, 5, 6, 7, 8,
        9,10,11,12,13,14,15,16, 0,17,
       18,19,20,21,22,23,24,25,26,27,0)
   );
var
   Jamo        : array[1..3] of byte;
   FontNumber  : array[1..3] of integer;
   i, j        : integer;
   Order       : integer;
   Templet     : byte;
   TempBuffer  : THangulBuffer;
   Buffer      : THangulBuffer;

 procedure Color2RGB(Color : Longint; var R, G, B : byte);
 begin
    R := Byte(Color);
    G := Byte(Color shr 8);
    B := Byte(Color shr 16);
 end;

 procedure RGB2Color(R, G, B : byte; var Color : Longint);
 begin
    Color := (Longint(B) shl 16) or (Longint(G) shl 8) or Longint(R);
 end;

 procedure PutUnitHangul(x,y : integer; Color : Longint);
 var
    i, j    : integer;
    R, G, B : byte;
 begin

    case HangulUnit of

       huPixel :
       for j := 0 to 255 do begin
          if (Buffer[j shr 3] and ($80 shr (j and $07))) > 0 then
//             putPixel_16bit(x+((j shr 3) and $01) shl 3 + (j and $07),y + (j shr 4),Color{+ j shr 5});
             PutPixel(x+((j shr 3) and $01) shl 3 + (j and $07),y + (j shr 4),Color{+ j shr 5});
       end;

       huGradation :
       begin
          Color2RGB(Color,R,G,B);
          for j := 0 to 15 do begin
             RGB2Color(integer(R) * (Gradient[1]-j) div Gradient[1],
                       integer(G) * (Gradient[2]-j) div Gradient[2],
                       integer(B) * (Gradient[3]-j) div Gradient[3], Color);
             for i := 0 to 15 do begin
                if (Buffer[j * 2 + i div 8] and ($80 shr (i and $07))) > 0 then
                   PutPixel(x+i,y+j,Color);
             end;
		  end;
       end;

    end;

 end;

begin

   if HangulMode = hmWansung then s := WS2JH.ConvertWS2JH(s);

   Order := 1;

   while Order <= Length(s) do begin

      if ((ord(s[Order]) and $80) > 0) and (Order < length(s)) then begin

         FillChar(FontNumber,sizeof(FontNumber),0);

         Jamo[1] :=  (ord(s[Order  ]) and $7C) shr 2;
         Jamo[2] := ((ord(s[Order  ]) and $03) shl 3) + (ord(s[Order+1]) shr 5);
         Jamo[3] :=  (ord(s[Order+1]) and $1F);

         for i := 1 to 3 do begin
            if Jamo[i] in [0..30] then Jamo[i] := ConvertTable[i,Jamo[i]]
                                  else Jamo[i] := 0;
         end;

         if Jamo[3] > 0 then begin

            if Jamo[2] in [1..8,21]            then FontNumber[1] := 95;
            if Jamo[2] in [9,13,14,18,19]      then FontNumber[1] := 114;
            if Jamo[2] in [10..12,15..17,20]   then FontNumber[1] := 133;
            if Jamo[1] in [1,16]               then FontNumber[2] := 194;
            if Jamo[1] in [2..15,17..19]       then FontNumber[2] := 215;
            if Jamo[2] in [1,3,10]             then FontNumber[3] := 236;
            if Jamo[2] in [5,7,12,15,17,20,21] then FontNumber[3] := 263;
            if Jamo[2] in [2,4,6,8,11,16]      then FontNumber[3] := 290;
            if Jamo[2] in [9,13,14,18,19]      then FontNumber[3] := 317;

         end else begin

            if Jamo[2] in [1..8,21]            then FontNumber[1] := 0;
            if Jamo[2] in [9,13,19]            then FontNumber[1] := 19;
            if Jamo[2] in [14,18]              then FontNumber[1] := 38;
            if Jamo[2] in [10..12,20]          then FontNumber[1] := 57;
            if Jamo[2] in [15..17]             then FontNumber[1] := 76;
            if Jamo[1] in [1,16]               then FontNumber[2] := 152;
			if Jamo[1] in [2..15,17..19]       then FontNumber[2] := 173;

            FontNumber[3] := 0;
         end;

         Move(HangulFontData^[FontNumber[1]+Jamo[1]-1],Buffer,32);

         Buffer[32] := 0;
         Buffer[33] := 0;

         if Jamo[2] > 0 then
         for i := 0 to 31 do begin
            Buffer[i] := Buffer[i] or HangulFontData^[FontNumber[2]+Jamo[2]-1][i div 2][i mod 2];
         end;

         if Jamo[3] > 0 then
         if FontNumber[3] > 0 then begin
            for i := 0 to 31 do begin
               Buffer[i] := Buffer[i] or HangulFontData^[FontNumber[3]+Jamo[3]-1][i div 2][i mod 2];
            end;
		 end;

		 if IsBold then begin

            TempBuffer := Buffer;

            for j :=  0 to     31 do Buffer[j] := Buffer[j] or  Buffer[j+2];
            for j := 31 downto  2 do Buffer[j] := Buffer[j] or  Buffer[j-2];
            for j :=  0 to     31 do Buffer[j] := Buffer[j] or (Buffer[j] shr 1);
            for j := 31 downto  0 do Buffer[j] := Buffer[j] or (Buffer[j] shl 1);

            putUnitHangul(x+(Order-1)*8,y,0);
            Buffer := TempBuffer;

         end;

         PutUnitHangul(x+(Order-1)*8,y,Color);
		 Inc(Order,2);

      end
      else
      if Assigned(AsciiFontData) then begin


         Templet := ord(s[Order]);
         if (Templet <  30) then Templet := 31;
         if (Templet > 126) then Templet := 32;
         Dec(Templet,31);

		 FillChar(Buffer,sizeof(Buffer),0);

		 for i := 0 to 15 do begin
			Buffer[i*2] := AsciiFontData^[Templet][i];
		 end;

         if IsBold then begin

            TempBuffer := Buffer;

            for j :=  0 to     15 do Buffer[j*2] := Buffer[j*2] or  Buffer[(j+1)*2];
            for j := 15 downto  1 do Buffer[j*2] := Buffer[j*2] or  Buffer[(j-1)*2];
            for j :=  0 to     15 do Buffer[j*2] := Buffer[j*2] or (Buffer[j*2] shr 1);
            for j := 15 downto  0 do Buffer[j*2] := Buffer[j*2] or (Buffer[j*2] shl 1);

            putUnitHangul(x+(Order-1)*8,y,0);
            Buffer := TempBuffer;
         end;

         PutUnitHangul(x+(Order-1)*8,y,Color);
         Inc(Order);
      end
      else begin
         Inc(Order);
      end;
   end;
end;

end.
